/* $Id: tds.c,v 1.108 2002/03/18 10:06:49 rio Exp $ */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/types.h>
#include <sys/conf.h>
#include <sys/errno.h>
#include <sys/bus.h>
#include <sys/module.h>
#include <sys/kernel.h>
#include <sys/uio.h>
#include <sys/time.h>

#include <machine/bus_pio.h>
#include <machine/bus.h>

#include <sys/rman.h>
#include <machine/resource.h>

#include <pci/pcivar.h>
#include <pci/pcireg.h>

#include <limits.h>

#include <pci/tdsvar.h>

#if !defined(lint)
static const char rcsid[]
	= "$Id: tds.c,v 1.108 2002/03/18 10:06:49 rio Exp $";
#endif


#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef	BITSPERBYTE
#define BITSPERBYTE	8
#endif

#define TDS_INTR_SET_PIO_MODE	(TDS_INTR_DATA_EXIST | TDS_INTR_DATA_FULL)

#define TDS_INTR_SET_DMA_MODE	(TDS_INTR_DMA_DONE | TDS_INTR_DATA_FULL)

enum tds_bool {
    tds_true = 1,
    tds_false = 0
};


enum tds_debug {
    tds_none = 0,
    tds_info = 1,
    tds_debug = 2,
    tds_more_debug = 3
};


#ifdef TDS_DEBUG

#if TDS_DEBUG > 0
#define TDS_DEBUG_DEFAULT	TDS_DEBUG
#else
#define TDS_DEBUG_DEFAULT	1
#endif

#if TDS_DEBUG < 1
#define TDS_INLINE_DEBUG	__inline
#else
#define TDS_INLINE_DEBUG	/**/
#endif

#if TDS_DEBUG < 2
#define TDS_INLINE	__inline
#else
#define TDS_INLINE	/**/
#endif

#define STATIC		/**/
#define	TDS_INFINITE	(300 * hz)

STATIC int tds_debug_level = TDS_DEBUG_DEFAULT;

#define ASSERT(cond)	KASSERT(cond, #cond)
#define IFDBG(level,cond)	if(((level) <= tds_debug_level) && (cond))
#define DMSG(level,fmt,args...) \
	do { if((level) <= tds_debug_level) printf(fmt, ##args); } while(0)

#else	/* TDS_DEBUG */

#define TDS_DEBUG_DEFAULT	0

#define STATIC		static
#define TDS_INLINE	__inline
#define	TDS_INFINITE	0

#define ASSERT(cond)	/*none*/
#define IFDBG(level,cond)	if(0)
#define DMSG(level,fmt,args...) \
	do { } while(0)

#endif	/* TDS_DEBUG */


/* driver name */
static char tds_driver_name[] = "tds";


static struct tds_type tds_devs[] = {
    { NITSUKI_VENDERID,
      NITSUKI_TIME_SYNC_DATA_SAMPLER,
      "Nitsuki Time Sync Data Sampler" },
    { 0, 0, NULL },
};


static int	tds_probe		__P((device_t));
static int	tds_attach		__P((device_t));
static int	tds_detach		__P((device_t));
static int	tds_shutdown	__P((device_t));
void		tds_intr		__P((void *));


static device_method_t tds_methods[] = {
    DEVMETHOD(device_probe,	tds_probe),
    DEVMETHOD(device_attach,	tds_attach),
    DEVMETHOD(device_detach,	tds_detach),
    DEVMETHOD(device_shutdown,	tds_shutdown),
    DEVMETHOD(bus_print_child,	bus_generic_print_child),
    DEVMETHOD(bus_driver_added,	bus_generic_driver_added),
    { 0, 0 },
};


static driver_t tds_driver = {
    tds_driver_name,
    tds_methods,
    sizeof(struct tds_softc),
};


static devclass_t tds_devclass;


DRIVER_MODULE(tds, pci, tds_driver, tds_devclass, 0, 0);


static int	tdsopen		__P((dev_t, int, int, struct proc *));
static int	tdsclose	__P((dev_t, int, int, struct proc *));
static int	tdsread		__P((dev_t, struct uio *, int));
static int	tdswrite	__P((dev_t, struct uio *, int));
static int	tdsioctl	__P((dev_t, u_long, caddr_t, int, struct proc *));



static struct cdevsw tds_cdevsw = {
    tdsopen,
    tdsclose,
    tdsread,
    tdswrite,
    tdsioctl,
    nopoll,		/* poll		*/
    nommap,		/* mmap		*/
    nostrategy,		/* strategy	*/
    tds_driver_name,
    TDS_MAJOR,
    nodump,		/* dump		*/
    nopsize,		/* psize	*/
    0,			/* flags	*/
    -1			/* bmaj		*/
};


/* 2004.03 K.Kawano DMA Mode	*/
static u_int8_t  dma_mode;

/* device register handling */
#define	TDS_IOPORT(off)	((u_int) (sc->iobase + off))


static u_int8_t tds_dbuf_clr_seq[] = { 3, 2, 0, };

static TDS_INLINE void
tdsreg_clear_fifo(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_FIFO_CLEAR);
	int num = sizeof(tds_dbuf_clr_seq) / sizeof(u_int8_t);
	int i;
	for (i = 0; i < num; i ++) {
		outb(reg, tds_dbuf_clr_seq[i]);
		}
}


#ifdef	OBSOLATED
static TDS_INLINE void
tdsreg_check_signal(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_SYNCRONIZE);
	outb(reg, TDS_SIGNAL_CHECK);
}
#endif	/*OBSOLATED*/


static TDS_INLINE void
tdsreg_sync_1pps(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_SYNCRONIZE);
	outb(reg, TDS_1PPS_SYNC);
	outb(reg, 0);
}


static TDS_INLINE void
tdsreg_sync_time(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_SYNCRONIZE);
	outb(reg, TDS_TIME_SYNC);
}


static TDS_INLINE void
tdsreg_set_irq(struct tds_softc *sc, u_int8_t flags)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_INTERRUPT_SET);
	IFDBG(tds_more_debug,1) {
		device_printf(sc->devicet, "Set Interrupt (0x%02x)\n", flags);
		}
	outb(reg, flags);
}


static TDS_INLINE u_int8_t
tdsreg_get_irq(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_READ_INTERRUPT);
	u_int8_t val = inb(reg);
	return val;
}


static TDS_INLINE void
tdsreg_set_time(struct tds_softc *sc, u_int32_t time)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_TIME_SET);
	outl(reg, time);
}


static TDS_INLINE u_int32_t
tdsreg_get_time(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_READ_TIME);
	u_int32_t val = inl(reg);
	return val;
}


static TDS_INLINE u_int8_t
tdsreg_get_year(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_READ_YEAR);
	u_int8_t val = inb(reg);
	return val;
}


static TDS_INLINE void
tdsreg_set_sampling(struct tds_softc *sc, u_int8_t flags)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_SAMPLING_SET);
	outb(reg, flags);
}


static TDS_INLINE u_int8_t
tdsreg_get_status(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_READ_STATUS);
	u_int8_t val = inb(reg);
	return val;
}


static TDS_INLINE void
tdsreg_start_sampling(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_SAMPLING_START);
	outb(reg, TDS_SAMPLING_START);
}


static TDS_INLINE void
tdsreg_stop_sampling(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_SAMPLING_START);
	outb(reg, TDS_SAMPLING_STOP);
}


static TDS_INLINE u_int32_t
tdsreg_get_data(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_READ_DATA);
	u_int32_t val = inl(reg);
	return val;
}


static TDS_INLINE u_int8_t
tdsreg_get_input_status(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_READ_INPUT);
	u_int8_t val = inb(reg);
	return val;
}


static TDS_INLINE void
tdsreg_set_dma_address(struct tds_softc *sc, u_int32_t dstaddr)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_DMA_ADDR_SET);
	outl(reg, dstaddr);
}


static TDS_INLINE void
tdsreg_set_dma_datasize(struct tds_softc *sc, u_int32_t dsize)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_DMA_SIZE_SET);
	outl(reg, dsize);
}


static TDS_INLINE u_int32_t
tdsreg_get_dma_datasize(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_READ_DMA_SIZE);
	u_int32_t val = inl(reg);
	return val;
}


static TDS_INLINE void
tdsreg_start_dma(struct tds_softc *sc)
{
	/* 2004.05 K.Kawano	*/
	if ((dma_mode & TDS_DMA_MODE_HARD) == 0) {
		u_int reg = TDS_IOPORT(TDS_REG_WRITE_DMA_START);
		outb(reg, TDS_DMA_START);
		}
	else {
		u_int reg = TDS_IOPORT(TDS_REG_WRITE_DMA_MODE);
		outb(reg, dma_mode);
		}
}

static TDS_INLINE void
tdsreg_stop_dma(struct tds_softc *sc)
{
	if ((dma_mode & TDS_DMA_MODE_HARD) == 0) {
		u_int reg = TDS_IOPORT(TDS_REG_WRITE_DMA_START);
		outb(reg, TDS_DMA_STOP);
		}
	else {
		u_int reg = TDS_IOPORT(TDS_REG_WRITE_DMA_MODE);
		outb(reg, TDS_DMA_MODE_SOFT);
		dma_mode = TDS_DMA_MODE_SOFT;
		}
}


static TDS_INLINE u_int8_t
tdsreg_get_dma_status(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_READ_DMA_STATUS);
	return inb(reg);
}


static TDS_INLINE void
tdsreg_set_dma_mode(struct tds_softc *sc, u_int8_t mode)
{
	/* 2004.05 K.Kawano	*/
/*
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_DMA_MODE);
	outb(reg, mode);
*/
	dma_mode = mode;
}


static TDS_INLINE u_int8_t
tdsreg_get_dma_mode(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_READ_DMA_MODE);
	return inb(reg);
}


static TDS_INLINE u_int32_t
tdsreg_get_write_counter(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_READ_WRITE_COUNTER);
	u_int32_t val  = inl(reg);
	return val;
}


static TDS_INLINE u_int32_t
tdsreg_get_read_counter(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_READ_READ_COUNTER);
	u_int32_t val  = inl(reg);
	return val;
}


static TDS_INLINE void
tdsreg_clear_write_counter(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_CLEAR_WRITE_COUNTER);
	outb(reg, TDS_FIFO_COUNTER_CLEAR);
}


static TDS_INLINE void
tdsreg_clear_read_counter(struct tds_softc *sc)
{
	u_int reg = TDS_IOPORT(TDS_REG_WRITE_CLEAR_READ_COUNTER);
	outb(reg, TDS_FIFO_COUNTER_CLEAR);
}



/* support functions */

static int
tds_wait_next_1pps(struct tds_softc *sc)
{
	u_int8_t saved_intr = sc->intr_flag;
	intrmask_t plev;
	int rv = 0;

	plev = spltds();
	sc->sync_1pps_count = 0;
	sc->intr_flag |= TDS_INTR_1PPS;
	tdsreg_set_irq(sc, sc->intr_flag);
	while(sc->sync_1pps_count == 0) {
		rv = tsleep(&sc->sync_1pps_count, TDSPRI | PCATCH, "1PPS", TDS_INFINITE);
		if (rv != 0) break;
		}
	sc->intr_flag = saved_intr;
	tdsreg_set_irq(sc, saved_intr);
	splx(plev);
	if (rv != 0) {
		device_printf(sc->devicet, "tsleep() failed (wait for 1pps)\n");
		sc->error = TDS_ERROR_UNKNOWN;
		}
	return rv;
}

static TDS_INLINE void
tds_issue_dma(struct tds_softc *sc)
{
	int idx = sc->dmabuf_write;
	tds_dma_buf_t *buf = &(sc->dmabuf[idx]);

	tdsreg_set_dma_mode(sc, TDS_DMA_MODE_HARD | TDS_DMA_MODE_RELATIVE);
	tdsreg_set_dma_datasize(sc, sc->dma_chunk_size / sizeof(u_int32_t));
	tdsreg_set_dma_address(sc, (u_int32_t)buf->physaddr);
	tdsreg_start_dma(sc);
	device_printf(sc->devicet,
	  "DMA start (%d word) with buffer[%d] phys 0x%lx\n",
	  sc->dma_chunk_size / sizeof(u_int32_t),
	  idx, (unsigned long)buf->physaddr);
	return;
}


static TDS_INLINE int
tds_check_extension_card(struct tds_softc *sc)
{
	u_int8_t stat;
	int rv;

	stat = tdsreg_get_input_status(sc);
	if (stat & TDS_EXTCARD_EXIST) {
		rv = tds_true;
		}
	else {
		rv = tds_false;
		}
	return rv;
}


static void
tds_dma_setmap(void *arg, bus_dma_segment_t *segs, int nseg, int error)
{
	tds_dma_buf_t *buf = (tds_dma_buf_t *)arg;
	struct tds_softc *sc = buf->sc;

	if (error == 0) {
		buf->physaddr = (void *)segs->ds_addr;
		}
	else {
		buf->physaddr = 0;
		}

	device_printf(sc->devicet,
		"tds_dma_setmap(VA=%lx, PA=%lx, SZ=%lx), nseg=%d, error=%d\n",
		(unsigned long)buf->virtaddr,
		(unsigned long)segs->ds_addr,
		(unsigned long)segs->ds_len,
		nseg, error);
}

static TDS_INLINE void
tds_dma_buffer_init(struct tds_softc *sc)
{
	int i;

	for (i = 0; i < TDS_DMA_NSEG; i ++) {
		tds_dma_buf_t *buf = &sc->dmabuf[i];
		buf->write_offset = 0;
		buf->read_offset = 0;
		buf->flags = 0;
#ifdef TDS_DEBUG
		bzero(buf->virtaddr, sc->dma_buffer_size);
#endif
	    }
	sc->dmabuf_write = 0;
	sc->dmabuf_read = 0;

	tds_issue_dma(sc);
}


static TDS_INLINE void
tds_calc_dma_chunk_size(struct tds_softc *sc)
{
#if (TDS_DMA_USE_CHUNK==0)
	/* bps: bits per sec */
	int bps = sc->sampling_freq * sc->sampling_width * sc->sampling_channel;
	/* we use chunk size 32KWord per 1Mbps */
	size_t size = 32 * bps / 1024;
	size_t chunk;

	for(chunk = TDS_DMA_MIN_CHUNK; size > chunk; chunk *= 2)	continue;
	if (chunk > TDS_DMA_MAX_CHUNK)	chunk = TDS_DMA_MAX_CHUNK;
	sc->dma_chunk_size = chunk * sizeof(u_int32_t);
#else	/*(TDS_DMA_USE_CHUNK==0)*/
	sc->dma_chunk_size = TDS_DMA_USE_CHUNK * sizeof(u_int32_t);
#endif	/*(TDS_DMA_USE_CHUNK==0)*/
	IFDBG(tds_info,1) {
		device_printf(sc->devicet,
		  "\tDMA chunk_size = 0x%05x(word)\n",
		  sc->dma_chunk_size/sizeof(u_int32_t));
		}
}

static int
tds_sampling_condition(struct tds_softc *sc, u_int32_t cond)
{
	int rv = 0;
	int freq = 0;
	int width = 0;
	int channel = 0;

	switch (cond & TDS_MASK_SAMP_FREQ) {
		case TDS_SAMPLING_40KHZ:
			freq = 40 * 1000;
			break;
		case TDS_SAMPLING_100KHZ:
			freq = 100 * 1000;
			break;
		 case TDS_SAMPLING_200KHZ:
			freq = 200 * 1000;
			break;
		case TDS_SAMPLING_500KHZ:
			freq = 500 * 1000;
			break;
		case TDS_SAMPLING_1MHZ:
			freq = 1 * 1000000;
			break;
		case TDS_SAMPLING_2MHZ:
			freq = 2 * 1000000;
			break;
		case TDS_SAMPLING_4MHZ:
			freq = 4 * 1000000;
			break;
		case TDS_SAMPLING_8MHZ:
			freq = 8 * 1000000;
			break;
		case TDS_SAMPLING_16MHZ:
			freq = 16 * 1000000;
			break;
		default:
			device_printf(sc->devicet, "unsupported freqency\n");
			sc->error = TDS_ERROR_NOT_SUPPORTED;
			rv = ENODEV;
			break;
		}
	switch (cond & TDS_MASK_SAMP_WIDTH) {
		case TDS_SAMPLING_1BIT:
			width = 1;
			break;
		case TDS_SAMPLING_2BIT:
			width = 2;
			break;
		case TDS_SAMPLING_4BIT:
			width = 4;
			break;
		case TDS_SAMPLING_8BIT:
			width = 8;
			break;
		}
	switch (cond & TDS_MASK_SAMP_CH) {
		case TDS_SAMPLING_1CH:
			channel = 1;
			break;
		case TDS_SAMPLING_4CH:
			channel = 4;

			/* 2004.03 K.Kawano SET ERROR	*/
			if ((freq * width) > 32000000) {
				device_printf(sc->devicet, "over 128Mbps sampling is NOT supported\n");
				sc->error = TDS_ERROR_NOT_SUPPORTED;
				rv = ENXIO;
				}

			/* it requires extension card, check it out */
			if (tds_check_extension_card(sc) == tds_false) {
				device_printf(sc->devicet, "extension card is NOT installed\n");
				sc->error = TDS_ERROR_EXTCARD_NOT_EXIST;
#ifdef NOT_YET
				rv = ENXIO;
#endif
				}
			break;
		}
	if (rv == 0) { /* valid */
		sc->sampling_freq = freq;
		sc->sampling_width = width;
		sc->sampling_channel = channel;
		tds_calc_dma_chunk_size(sc);
		IFDBG(tds_info,1) {
			printf("%s: sampling condition %dHz/%dbit/%dch\n",
				sc->dev_name, freq, width, channel);
			}
		}
	return rv;
}

static TDS_INLINE int
tds_dma_buf_free(struct tds_softc *sc)
{
	if (sc->dmatag != NULL) {
		int index;
		for (index = TDS_DMA_NSEG - 1; index >= 0; index--) {
			tds_dma_buf_t *buf = &(sc->dmabuf[index]);
			if (buf->physaddr != 0) {
				bus_dmamap_unload(sc->dmatag, buf->dmamap);
				buf->physaddr = 0;
				}
			if (buf->virtaddr != 0) {
				bus_dmamem_free(sc->dmatag, buf->virtaddr, buf->dmamap);
				buf->virtaddr = 0;
				buf->dmamap = NULL;
				}
			}
		bus_dma_tag_destroy(sc->dmatag);
		}
	sc->dmatag = NULL;
	return 0;
}

static TDS_INLINE int
tds_dma_buf_alloc(struct tds_softc *sc)
{
	int err;
	int fail_count;
	const int fail_max = 3;
	int index;
	size_t dma_size = sc->dma_buffer_size * 2;

	for (index = 0; index < TDS_DMA_NSEG; index ++) {
		sc->dmabuf[index].sc = sc;
		}

	err = bus_dma_tag_create(NULL, PAGE_SIZE, 0, BUS_SPACE_MAXADDR_32BIT,
		BUS_SPACE_MAXADDR, NULL, NULL,
		dma_size, 1, dma_size,
		BUS_DMA_ALLOCNOW, &sc->dmatag);
	if (err != 0) {
		device_printf(sc->devicet, "bus_dma_tag_create() failure.\n");
		sc->dmatag = NULL;
		return (ENXIO);
		}

	fail_count = 0;
	index = 0;
	while ((index < TDS_DMA_NSEG) && (fail_count < fail_max)) {
		tds_dma_buf_t *buf = &(sc->dmabuf[index]);
		err = bus_dmamem_alloc(sc->dmatag, &(buf->virtaddr),BUS_DMA_NOWAIT, &(buf->dmamap) );
		if (err != 0) {
			device_printf(sc->devicet, "bus_dmamem_alloc() failure.\n");
			bus_dmamem_free(sc->dmatag, buf->virtaddr, buf->dmamap);
			buf->virtaddr = 0;
			fail_count++;
			}
		else {
			index++;
			}
		}
	if (fail_count >= fail_max) {
		device_printf(sc->devicet, "bus_dmamem_alloc() too many failure\n");
		tds_dma_buf_free(sc);
		return (ENXIO);
		}

	fail_count = 0;
	index = 0;
	while ((index < TDS_DMA_NSEG) && (fail_count < fail_max)) {
		tds_dma_buf_t *buf = &(sc->dmabuf[index]);
		if (buf->virtaddr != 0) {
			err = bus_dmamap_load(sc->dmatag, buf->dmamap, buf->virtaddr,dma_size, tds_dma_setmap, buf, 0);
			if ((err != 0) || (buf->physaddr == 0)) {
				device_printf(sc->devicet, "bus_dmamap_load() failure.\n");
				buf->physaddr = 0;
				fail_count++;
				}
			else {
				index++;
				}
			}
		else {
			device_printf(sc->devicet, "WARNING: buf->virtaddr == 0!!\n");
			}
		}
	if (fail_count >= fail_max) {
		device_printf(sc->devicet, "bus_dmamap_load() too many failure\n");
		tds_dma_buf_free(sc);
		return (ENXIO);
		}

	return 0;
}


static TDS_INLINE int
tds_init(struct tds_softc *sc, int mode)
{
	int rv = 0;

#ifdef TDS_DEBUG
	sc->verbose = tds_debug;
#else
	sc->verbose = tds_none;
#endif
	sc->error = TDS_ERROR_NO_ERROR;
	sc->read_mode = mode;
	sc->max_fifo_overflow = TDS_MAX_FIFO_OVERFLOW_DEFAULT;
	sc->sampling_flag = TDS_SAMPLING_DEFAULT;
	sc->intr_flag = 0;
	/*sc->is_open = 0;*/
	sc->sampling_done = 0;
	sc->pps_sync_done = 0;
	sc->time_sync_done = 0;
	sc->time_set_done = 0;
	sc->sampling_start = 0;
	sc->sync_1pps_count = 0;
	sc->sync_data_count = 0;
	sc->fifo_underflow_count = 0;
	sc->fifo_overhalf_count = 0;
	sc->fifo_overflow_count = 0;
	sc->dma_buffer_size = TDS_DMA_SIZE;
	sc->dma_chunk_size = TDS_DMA_MIN_CHUNK * sizeof(u_int32_t);	/* fail safe */

	switch (sc->read_mode) {
		case TDS_READ_USE_PIO:
			if (sc->dmatag != NULL) {
				rv = tds_dma_buf_free(sc);
				}
			break;
		case TDS_READ_USE_DMA:
			if (sc->dmatag == NULL) {
				rv = tds_dma_buf_alloc(sc);
				}
			break;
		default:
			panic("%s: unknown read mode (%d)\n", sc->dev_name, sc->read_mode);
			break;
		}

	if (tds_sampling_condition(sc, sc->sampling_flag) == 0) {
		tdsreg_set_sampling(sc, sc->sampling_flag);
		}

	return 0;
}


static TDS_INLINE int
tds_get_fifo_count(struct tds_softc *sc)
{
	int wcount = tdsreg_get_write_counter(sc);
	int rcount = tdsreg_get_read_counter(sc);

	if ( wcount < rcount )	wcount += 0x40000000;
	return (wcount - rcount);
}

static TDS_INLINE int
tdsio_sampling_start(struct tds_softc *sc)
{
	int rv = 0;
	u_int8_t status;

	/* 2004.05 K.Kawano ... Disp driver version	*/
	device_printf(sc->devicet, "%s\n",_szVer);

#ifdef	OBSOLATED
	tdsreg_check_signal(sc);
	rv = tds_wait_next_1pps(sc);
	if (rv != 0) {
		return rv;
		}
#endif	/*OBSOLATED*/

/* 2004.05 K.Kawano ... No Stop by Time Sync (or set)	*/
	if (sc->pps_sync_done == 0) {
		device_printf(sc->devicet, "not synchronized to 1pps yet\n");
		sc->error = TDS_ERROR_NOT_SYNC_1PPS;
		rv = EIO;
		}
	else if ((sc->time_sync_done == 0) && (sc->time_set_done == 0)) {
		device_printf(sc->devicet,"not synchronized (or set) time yet\n");
		sc->error = TDS_ERROR_NOT_SYNC_TIME;
		rv = EIO;
		}

	status = tdsreg_get_status(sc);
	if (!(status & TDS_STAT_10MHZ_INPUT)) {
		device_printf(sc->devicet, "10MHz input not detected\n");
		sc->error = TDS_ERROR_10MHZ_NOT_DETECT;
		rv = EIO;
	    }
	else {
		intrmask_t plev;
		switch (sc->read_mode) {
			case TDS_READ_USE_PIO:
				plev = spltds();
				sc->intr_flag = TDS_INTR_SET_PIO_MODE;
				tdsreg_set_irq(sc, sc->intr_flag);
				sc->sampling_start = 1;
				tdsreg_start_sampling(sc);
				splx(plev);
				break;
			case TDS_READ_USE_DMA:
				plev = spltds();
				tds_dma_buffer_init(sc);
				/* 2004.03 K.Kawano Intr Reg. Clear	*/
				status = tdsreg_get_irq(sc);

				sc->intr_flag = TDS_INTR_SET_DMA_MODE;
				tdsreg_set_irq(sc, sc->intr_flag);
				sc->sampling_start = 1;
				tdsreg_start_sampling(sc);
				splx(plev);
				/* actual sampling will start after next 1PPS sync */
				break;
			default:
				panic("%s: unknown read mode (%d)\n",sc->dev_name, sc->read_mode);
				break;
			}
		}
	return rv;
}

static TDS_INLINE int
tdsio_sampling_stop(struct tds_softc *sc)
{
	intrmask_t plev;
	switch(sc->read_mode) {
		case TDS_READ_USE_PIO:
			plev = spltds();
			sc->sampling_start = 0;
			tdsreg_stop_sampling(sc);
			splx(plev);
			break;
		case TDS_READ_USE_DMA:
			plev = spltds();
			tdsreg_stop_dma(sc);
/*
			tdsreg_set_dma_mode(sc, TDS_DMA_MODE_SOFT);
*/
			sc->sampling_start = 0;
			tdsreg_stop_sampling(sc);
			/* wait for next 1PPS sync intrrupt before disabling DMA */
			sc->sync_1pps_count = 0;
			sc->intr_flag = TDS_INTR_SET_PIO_MODE | TDS_INTR_1PPS;
			tdsreg_set_irq(sc, sc->intr_flag);
			while(sc->sync_1pps_count == 0) {
				tsleep(&sc->sync_1pps_count, TDSPRI | PCATCH, "1PPS", TDS_INFINITE);
				}

			/* 2004.03 K.Kawano add  Wait +1sec	*/
/*
			sc->intr_flag &= ~TDS_INTR_DATA_FULL;
			tdsreg_set_irq(sc, sc->intr_flag);
			tdsreg_set_dma_mode(sc, TDS_DMA_MODE_SOFT);
			sc->sync_1pps_count = 0;
			while(sc->sync_1pps_count == 0) {
				tsleep(&sc->sync_1pps_count, TDSPRI | PCATCH, "1PPS", TDS_INFINITE);
				}
			sc->sampling_start = 0;
*/

			sc->intr_flag = 0;
			tdsreg_set_irq(sc, sc->intr_flag);
			splx(plev);
			break;
		default:
			panic("%s: unknown read mode (%d)\n",sc->dev_name, sc->read_mode);
			break;
		}
	return 0;
}


/* driver routine */

static int
tds_probe(device_t dev)
{
	struct tds_type *tt = tds_devs;
	u_int16_t vid = pci_get_vendor(dev);
	u_int16_t did = pci_get_device(dev);

	while (tt->tds_name != NULL) {
		if ((vid == tt->tds_vid) && (did == tt->tds_did)) {
			device_set_desc(dev, tt->tds_name);
			return (0);
			}
		tt++;
		}

	return ENXIO;
}


static int
tds_attach(device_t dev)
{
	struct tds_softc	*sc;
	int			rid;
	int			err;

	sc = device_get_softc(dev);
	bzero(sc, sizeof(*sc));

	sc->unit = device_get_unit(dev);
	sc->devt = makedev(TDS_MAJOR, sc->unit);
	sc->devicet = dev;
	sc->dev_str = tds_driver_name;
	sprintf(sc->dev_name, "%s%d", sc->dev_str, sc->unit);

	rid = PCI_MAP_REG_START;

	sc->res_io =
	bus_alloc_resource(dev, SYS_RES_IOPORT, &rid, 0, ~0, 1, RF_ACTIVE);

	if (sc->res_io == NULL) {
		device_printf(dev, "Could not map I/O port\n");
		return ENXIO;
		}
	sc->iobase = rman_get_start(sc->res_io);

	make_dev(&tds_cdevsw, sc->unit, UID_ROOT, GID_WHEEL, 0600,"%s", sc->dev_name);

	rid = 0;
	sc->res_irq = bus_alloc_resource(dev, SYS_RES_IRQ, &rid, 0ul, ~0ul, 1, RF_SHAREABLE | RF_ACTIVE);
	if (sc->res_irq) {
		bus_setup_intr(dev, sc->res_irq, INTR_TYPE_TDS, tds_intr, sc, &sc->cookie);
		}
	else {
		device_printf(dev, "Could not map interrupt\n");
		bus_release_resource(dev,SYS_RES_IOPORT,PCI_MAP_REG_START,sc->res_io);
		return ENXIO;
		}

	sc->rev = pci_get_revid(dev);
	sc->irq = pci_get_irq(dev);

	{
		u_int32_t data;
		data = pci_read_config(dev, PCIR_COMMAND, 2);
		if ((sc->verbose >= tds_info) || bootverbose) {
			device_printf(dev, "\tCommand: %04xh\n", data);
			}
		data |= (PCIM_CMD_PORTEN | PCIM_CMD_BUSMASTEREN);
		pci_write_config(dev, PCIR_COMMAND, data, 2);
		data = pci_read_config(dev, PCIR_COMMAND, 2);
		if ((sc->verbose >= tds_info) || bootverbose) {
			device_printf(dev, "\tCommand: %04xh\n", data);
			}
	}

	err = tds_init(sc, TDS_READ_DEFAULT);
	if (err != 0) {
		bus_teardown_intr(dev, sc->res_irq, sc->cookie);
		bus_release_resource(dev, SYS_RES_IRQ, 0, sc->res_irq);
		destroy_dev(sc->devt);
		bus_release_resource(dev, SYS_RES_IOPORT, PCI_MAP_REG_START, sc->res_io);
		}

	sc->intr_flag = 0;
	tdsreg_set_irq(sc, sc->intr_flag);
	tdsreg_stop_sampling(sc);

	if ((sc->verbose >= tds_info) || bootverbose) {
		device_printf(dev, "\tClass: %02xh\n", pci_get_class(dev));
		device_printf(dev, "\tSubClass: %02xh\n", pci_get_subclass(dev));
		device_printf(dev, "\tProgramming I/F: %02xh\n", pci_get_progif(dev));
		device_printf(dev, "\tRevision: %d\n", sc->rev);
		device_printf(dev, "\tBus No: %d\n", pci_get_bus(dev));
		device_printf(dev, "\tDevice No: %d\n", pci_get_slot(dev));
		device_printf(dev, "\tFunction No: %d\n", pci_get_function(dev));
		device_printf(dev, "\tBase I/O Address: %08xh\n", sc->iobase);
		device_printf(dev, "\tIRQ: %d\n", sc->irq);
	    }

	return 0;
}


static int
tds_detach(device_t dev)
{
	struct tds_softc *sc;
	dev_t devt;
	sc = device_get_softc(dev);
	devt = sc->devt;

	tdsreg_stop_sampling(sc);
	tds_dma_buf_free(sc);

	bus_teardown_intr(dev, sc->res_irq, sc->cookie);
	bus_release_resource(dev, SYS_RES_IRQ, 0, sc->res_irq);
	destroy_dev(devt);
	bus_release_resource(dev, SYS_RES_IOPORT, PCI_MAP_REG_START, sc->res_io);

	return 0;
}


static int
tds_shutdown(device_t dev)
{
	/* XXX: do I have to do anything ? */
	return 0;
}


void
tds_intr(void *arg)
{
	struct tds_softc *sc = (struct tds_softc *) arg;
	u_int8_t intr;
	u_int8_t save_intr;

	IFDBG(tds_more_debug,1) {
		device_printf(sc->devicet, "called tds_intr()\n");
		}

	save_intr = sc->intr_flag;

	while((intr = (tdsreg_get_irq(sc) & sc->intr_flag)) != 0) {
		IFDBG(tds_more_debug,1) {
			struct timeval time;
			getmicrotime(&time);
			device_printf(sc->devicet, "Get interrupt (0x%02x) at %ld.%ld\n",
			  (int)intr, time.tv_sec, time.tv_usec);
			}
		if (intr & TDS_INTR_DATA_HALF) {
			IFDBG(tds_more_debug,1) {
				struct timeval time;
				getmicrotime(&time);
				device_printf(sc->devicet, "FIFO half full at %ld.%ld\n", time.tv_sec, time.tv_usec);
				}
			sc->fifo_overhalf_count++;
			/* wakeup(&sc->fifo_overhalf_count); */
			}
		if (intr & TDS_INTR_DATA_FULL) {
			struct timeval time;

			/* 2004.05 K.Kawano Check	*/
/*
			sc->intr_flag = 0;
			tdsreg_set_irq(sc, sc->intr_flag);
			tdsreg_clear_fifo(sc);
			device_printf(sc->devicet,"request FIFO clear\n");
*/
			getmicrotime(&time);
			device_printf(sc->devicet, "FIFO full at %ld.%ld\n", time.tv_sec, time.tv_usec);

			/* 2004.03 K.Kawano Check	*/
			tdsreg_clear_fifo(sc);
			device_printf(sc->devicet,"request FIFO clear\n");
			sc->fifo_overflow_count++;
/*
			if (sc->sampling_start) {
				sc->sampling_start = 0;
				tdsreg_stop_sampling(sc);
				if (sc->intr_flag != TDS_INTR_SET_PIO_MODE) {
					sc->intr_flag = TDS_INTR_SET_PIO_MODE;
					tdsreg_set_irq(sc, sc->intr_flag);
					wakeup(&sc->wait_buffer_ready);
					}
				else {
					wakeup(&sc->sync_data_count);
					}
				}
*/
			}
		if (intr & TDS_INTR_DMA_DONE) {
			int idx = sc->dmabuf_write;
			tds_dma_buf_t *buf = &(sc->dmabuf[idx]);
			u_int32_t size = tdsreg_get_dma_datasize(sc) * sizeof(u_int32_t);
			buf->write_offset += size;
			/* prepare next DMA buffer to avoid overrrun A.S.A.P. */
			if (buf->write_offset + sc->dma_chunk_size >= sc->dma_buffer_size) {
				int nextidx = (idx + 1) % TDS_DMA_NSEG;
				tds_dma_buf_t *nextbuf = &(sc->dmabuf[nextidx]);
				buf->flags |= TDS_BUFFLAG_DATA_READY;
				if (nextbuf->flags == 0) {
					/* next buffer is free, available for DMA target */
					tdsreg_set_dma_address(sc, (u_int32_t)nextbuf->physaddr);
					nextbuf->flags |= TDS_BUFFLAG_WRITE_BUSY;
					sc->dmabuf_write = nextidx;
					tdsreg_start_dma(sc);
					}
				else {
					/* next buffer is NOT free, need to stop DMA */
					tdsreg_stop_dma(sc);
/* 2004.05 K.Kawano	*/
/*					tdsreg_set_dma_mode(sc, TDS_DMA_MODE_SOFT);
*/
					sc->sampling_start = 0;
					tdsreg_stop_sampling(sc);
					sc->intr_flag = TDS_INTR_SET_PIO_MODE;
					tdsreg_set_irq(sc, sc->intr_flag);

					device_printf(sc->devicet, "DMA buffer full\n");
					}
				}
			else {
				/* Enable next chunk of DMA. */
				tdsreg_start_dma(sc);
				}
			if (size > sc->dma_chunk_size) {
				device_printf(sc->devicet,
				  "DMA finished words %d > %d, at buff[%d] offset %d\n",
				  size / sizeof(u_int32_t),
				  sc->dma_chunk_size / sizeof(u_int32_t),
				  idx, buf->write_offset);
				}
			IFDBG(tds_more_debug,1) {
				struct timeval time;
				getmicrotime(&time);
				device_printf(sc->devicet,
				  "DMA end (%d word) with buffer[%d] at %ld.%ld\n",
				  size/sizeof(u_int32_t), idx, time.tv_sec, time.tv_usec);
				}
			/* update current buf paramers available to reader */
			wakeup(&sc->wait_buffer_ready);
			}
		if (intr & TDS_INTR_DATA_EXIST) {
/* 2004.05 K.Kawano	*/
			if (sc->sampling_start == 0) {
				tdsreg_clear_fifo(sc);
				wakeup(&sc->wait_buffer_ready);
				}

			sc->sync_data_count++;
			wakeup(&sc->sync_data_count);
			}
		if (intr & TDS_INTR_1PPS) {
			sc->sync_1pps_count++;
			wakeup(&sc->sync_1pps_count);
			}
		}
}

static int
tdsopen(dev_t dev, int flag, int mode, struct proc *p)
{
	int			min;
	int			unit;
	struct tds_softc	*sc;

	min = minor(dev);
	unit = min & 0xff;
	sc = devclass_get_softc(tds_devclass, unit);
	IFDBG(tds_more_debug,1) {
		device_printf(sc->devicet, "called open()\n");
		}

	if ( sc->is_open == tds_true ) {
		return EBUSY;
		}

	sc->fifo_overflow_count = 0;
	sc->is_open = tds_true;

	return 0;
}


static int
tdsclose(dev_t dev, int flag, int mode, struct proc *p)
{
	int			min;
	int			unit;
	struct tds_softc	*sc;

	min = minor(dev);
	unit = min & 0xff;
	sc = devclass_get_softc(tds_devclass, unit);
	IFDBG(tds_more_debug,1) {
		device_printf(sc->devicet, "called close()\n");
		}

	if (sc->sampling_start) {
		(void)tdsio_sampling_stop(sc);
		}
	sc->is_open = tds_false;

	return 0;
}


static int
tdsread(dev_t dev, struct uio *uio, int flag)
{
	int			min;
	int			unit;
	struct tds_softc	*sc;
	intrmask_t		plev;
	int			rv = 0;

	min = minor(dev);
	unit = min & 0xff;
	sc = devclass_get_softc(tds_devclass, unit);
	IFDBG(tds_more_debug,1) {
		device_printf(sc->devicet, "called read()\n");
		}

	if (sc->sampling_done == 0) {
		device_printf(sc->devicet, "sampling flag is not set yet\n");
		sc->error = TDS_ERROR_NOT_SET_SAMPLING;
		return ENXIO;
		}

	if (sc->sampling_start == 0) {
		device_printf(sc->devicet, "sampling not started yet\n");
		sc->error = TDS_ERROR_NOT_STARTED;
		return EIO;
		}

	switch (sc->read_mode) {
		case TDS_READ_USE_PIO: {
			u_int32_t data;
			u_int8_t stat_flag;
			u_int8_t saved_stat;

			stat_flag = tdsreg_get_status(sc);
			if (stat_flag & TDS_STAT_FIFO_EMPTY) {
				/* sampling start with 1pps signal */
				tds_wait_next_1pps(sc);
				}

			saved_stat = stat_flag;
			while (uio->uio_resid > sizeof(u_int32_t)) {
				plev = spltds();
				stat_flag = tdsreg_get_status(sc);
				if (stat_flag != saved_stat) {
					IFDBG(tds_info,1) {
						device_printf(sc->devicet, "board status changed 0x%02x -> 0x%02x\n",
						  saved_stat, stat_flag);
						}
					saved_stat = stat_flag;
					}
				if (stat_flag & TDS_STAT_FIFO_EMPTY) {
					u_int8_t saved_intr = sc->intr_flag;
					u_int8_t intr = TDS_INTR_DATA_EXIST;
					sc->intr_flag = intr;
					tdsreg_set_irq(sc, intr);
					IFDBG(tds_info,1) {
						device_printf(sc->devicet, "IRQ setting changed 0x%02x -> 0x%02x\n",saved_intr,intr);
						device_printf(sc->devicet, "call tsleep() in read()\n");
						}
					rv = tsleep(&sc->sync_data_count,TDSPRI | PCATCH, "DATA", TDS_INFINITE);
					if (rv != 0) {
						device_printf(sc->devicet, "tsleep() failed (wait for DATA exist)\n");
						sc->error = TDS_ERROR_UNKNOWN;
						splx(plev);
						break;
						}
					/* XXX */
					sc->intr_flag = saved_intr;
					tdsreg_set_irq(sc, saved_intr);
					}
				splx(plev);
				if (stat_flag & TDS_STAT_FIFO_OVERFLOW) {
					device_printf(sc->devicet, "device FIFO overflow in read()\n");
					sc->fifo_overflow_count ++;
					if (sc->fifo_overflow_count > sc->max_fifo_overflow) {
						device_printf(sc->devicet, "too many times FIFO overflow");
						sc->error = TDS_ERROR_FIFO_OVERFLOW;
						rv = EIO;
						break;
						}
					}
				data = tdsreg_get_data(sc);
				rv = uiomove((caddr_t) &data, sizeof(data), uio);
				if (rv != 0) {
					device_printf(sc->devicet, "uiomove() returns %d\n", rv);
					sc->error = TDS_ERROR_UNKNOWN;
					break;
					}
				}
			if (uio->uio_resid > 0) {	/* deal fractions */
				/* XXX: */
				}
			break;
			}
		case TDS_READ_USE_DMA: {
			int *idx = &sc->dmabuf_read;
			while (uio->uio_resid > 0) {
				tds_dma_buf_t *buf = &(sc->dmabuf[*idx]);
				size_t size;
				caddr_t addr;

				/*
				 * block process, until data being available.
				 */
				plev = spltds();
				while(sc->dmabuf_read == sc->dmabuf_write
				  && buf->read_offset == buf->write_offset) {
					/* no data is available, wait for finishing DMA */
					rv = tsleep(&sc->wait_buffer_ready,TDSPRI | PCATCH, "READ", TDS_INFINITE);
	if (sc->sampling_start == 0) {
		rv = 1;
		device_printf(sc->devicet,"Exit Read Wait\n");
		}
					if (rv != 0) break;
					}
				splx(plev);
	if (sc->sampling_start == 0) {
		device_printf(sc->devicet, "FIFO Read Error\n");
		sc->error = TDS_ERROR_UNKNOWN;
		break;
		}
				if (rv != 0) {
					device_printf(sc->devicet, "tsleep() failed (wait for read buffer[%d])\n", *idx);
					sc->error = TDS_ERROR_UNKNOWN;
					break;
					}

				/*
				 * copy sampling data from DMA buffer to uio area.
				 */
				ASSERT(buf->write_offset > buf->read_offset);
				size = MIN(uio->uio_resid, buf->write_offset - buf->read_offset);
				addr = ((caddr_t)buf->virtaddr) + buf->read_offset;
				rv = uiomove(addr, size, uio);
				if (rv != 0) {
					device_printf(sc->devicet,"uiomove() returns %d\n", rv);
					sc->error = TDS_ERROR_UNKNOWN;
					break;
					}
				buf->read_offset += size;

				/*
				 * increment read_index, only when all data were read.
				 */
				plev = spltds();
				if (buf->read_offset == buf->write_offset
				  && (buf->flags & TDS_BUFFLAG_DATA_READY) != 0) {
					/* All data stored in currnet buffer were read */
					buf->read_offset = buf->write_offset = 0;
					buf->flags = 0;
					*idx = (*idx + 1) % TDS_DMA_NSEG;
					}
				splx(plev);
				}
			break;
			}
		default: {
			panic("%s: unknown read mode (%d)\n", sc->dev_name, sc->read_mode);
			break;
			}
		}

	return rv;
}


static int
tdswrite(dev_t dev, struct uio *uio, int flag)
{
	int		min;
	int		unit;
	struct tds_softc	*sc;

	min = minor(dev);
	unit = min & 0xff;
	sc = devclass_get_softc(tds_devclass, unit);
	IFDBG(tds_more_debug,1) {
		device_printf(sc->devicet, "called write()\n");
		}

	if ( sc->is_open != tds_true ) {
		return ENXIO;
		}

	device_printf(sc->devicet, "write operation not supported\n");
	sc->error = TDS_ERROR_NOT_SUPPORTED;
	return ENODEV;
}


static int
tdsioctl(dev_t dev, u_long cmd, caddr_t data, int flag, struct proc *p)
{
	int		min;
	int		unit;
	struct tds_softc	*sc;
	int rv = 0;

	min = minor(dev);
	unit = min & 0xff;
	sc = devclass_get_softc(tds_devclass, unit);
	IFDBG(tds_more_debug,1) {
		device_printf(sc->devicet,
		  "called ioctl(): command 0x%08x\n", (int)cmd);
		}

	switch (cmd) {
		case TDSIO_INIT:
			if (sc->sampling_start != 0) {
				device_printf(sc->devicet, "can't initialize while sampling in progress\n");
				sc->error = TDS_ERROR_ALREADY_STARTED;
				return EBUSY;
				}
			tds_init(sc, TDS_READ_DEFAULT);
			return 0;

		case TDSIO_BUFFER_CLEAR:
			if (sc->sampling_start != 0) {
				device_printf(sc->devicet, "can't clear FIFO while sampling in progress\n");
				sc->error = TDS_ERROR_ALREADY_STARTED;
				return EBUSY;
				}
			tdsreg_clear_fifo(sc);
			IFDBG(tds_info,1) {
				device_printf(sc->devicet, "cleared device FIFO\n");
				}
			return 0;

		case TDSIO_SYNC_1PPS: {
			int count = 0;
#ifdef	OBSOLATED
			tdsreg_check_signal(sc);
			rv = tds_wait_next_1pps(sc);
			if (rv != 0) {
			    break;
				}
#endif	/*OBSOLATED*/
			for (count = 0; count < 2; count++) {
				u_int8_t status = tdsreg_get_status(sc);
				if (status & TDS_STAT_1PPS_INPUT) {
					tdsreg_sync_1pps(sc);
					sc->pps_sync_done = 1;
					IFDBG(tds_info,1) {
						device_printf(sc->devicet, "synchronized to 1pps\n");
						}
					return 0;
					}
				/* XXX: one more try */
				rv = tds_wait_next_1pps(sc);
				if (rv != 0) {
					break;
					}
				}
			device_printf(sc->devicet, "1pps input not detected\n");
			sc->error = TDS_ERROR_1PPS_NOT_DETECT;
			return EIO;
			}
		case TDSIO_SYNC_TIME: {
			int count = 0;
#ifdef	OBSOLATED
			tdsreg_check_signal(sc);
			rv = tds_wait_next_1pps(sc);
			if (rv != 0) {
				break;
				}
#endif	/*OBSOLATED*/
			while (1) {
				u_int8_t status;
				count++;
				status = tdsreg_get_status(sc);
				if (status & TDS_STAT_TIME_INPUT) {
					tdsreg_sync_time(sc);
					rv = tds_wait_next_1pps(sc);
					sc->time_sync_done = 1;
					IFDBG(tds_info,1) {
						device_printf(sc->devicet, "synchronized to time\n");
						}
					break;
					}
				else {
					if (count > 2) {
						device_printf(sc->devicet, "time input not detected\n");
						sc->error = TDS_ERROR_TIME_NOT_DETECT;
						rv = EIO;
						break;
						}
					else {
						/* XXX: one more try */
						rv = tds_wait_next_1pps(sc);
						if (rv != 0) {
							break;
							}
						continue;
						}
					}
				}
			break;
			}
		case TDSIO_SET_INTERRUPT:	/* OBSOLETE */
		case TDSIO_GET_INTERRUPT:	/* OBSOLETE */
			break;	/* nothing to do */

		case TDSIO_SET_TIME: {
			u_int32_t *time = (u_int32_t *)data;
			tdsreg_set_time(sc, *time);
			sc->time_set_done = 1;
			IFDBG(tds_info,1) {
				device_printf(sc->devicet, "set time 0x%08x\n", *time);
				}
			break;
			}
		case TDSIO_GET_TIME: {
			u_int32_t *time = (u_int32_t *)data;
			*time = tdsreg_get_time(sc);
			break;
			}
		case TDSIO_GET_YEAR: {
			u_int8_t *year = (u_int8_t *)data;
			*year = tdsreg_get_year(sc);
			break;
			}
		case TDSIO_SET_SAMPLING: {
			u_int32_t *flags = (u_int32_t *)data;
			rv = tds_sampling_condition(sc, *flags);
			if (rv != 0) {
				break;
				}
			tdsreg_set_sampling(sc, *flags);
			sc->sampling_flag = *flags;
			sc->sampling_done = 1;
			break;
			}
		case TDSIO_GET_SAMPLING: {
			u_int32_t *flags = (u_int32_t *)data;
			if (sc->sampling_done == 0) {
				device_printf(sc->devicet, "not set sampling yet\n");
				sc->error = TDS_ERROR_NOT_SET_SAMPLING;
				rv = ENXIO;
				}
			else {
				*flags = sc->sampling_flag;
				}
			break;
			}
		case TDSIO_GET_STATUS: {
			u_int32_t *stat = (u_int32_t *)data;
			u_int8_t tmp;
			tmp = tdsreg_get_status(sc);
			*stat = (u_int32_t) tmp;
			break;
			}
		case TDSIO_SAMPLING_START:
			if (sc->sampling_start == 0) {
				rv = tdsio_sampling_start(sc);
				}
			else {
				device_printf(sc->devicet, "sampling already started\n");
				sc->error = TDS_ERROR_ALREADY_STARTED;
				rv = EBUSY;
				}
			break;

		case TDSIO_SAMPLING_STOP:
			if (sc->sampling_start != 0) {
				rv = tdsio_sampling_stop(sc);
				}
			else {
				device_printf(sc->devicet, "sampling not started yet\n");
				sc->error = TDS_ERROR_NOT_STARTED;
				rv = ENXIO;
				}
			break;

		case TDSIO_GET_ERROR: {
			int *error = (int *)data;
			*error = sc->error;
			break;
			}
		case TDSIO_SET_VERBOSE: {
			int *verb = (int *)data;
			sc->verbose = *verb;
			break;
			}
		case TDSIO_SET_MODE:	/* XXX: obsolete */
			/* nothing to do */
			break;

		case TDSIO_GET_MODE:	/* XXX: obsolete */
			/* nothing to do */
			break;

		case TDSIO_SET_ALLOW_OVERFLOW: {
			int *num = (int *)data;
			sc->max_fifo_overflow = *num;
			break;
			}
		case TDSIO_GET_OVERFLOW_COUNT: {
			int *num = (int *)data;
			*num = sc->fifo_overflow_count;
			break;
			}
		case TDSIO_SET_READ_MODE: {
			if (sc->sampling_start == 0) {
				int *mode = (int *)data;
				switch (*mode) {
					case TDS_READ_USE_PIO:
						sc->read_mode = *mode;
						if (sc->dmatag != NULL) {
							rv = tds_dma_buf_free(sc);
							}
						break;
					case TDS_READ_USE_DMA:
						sc->read_mode = *mode;
						if (sc->dmatag == NULL) {
							rv = tds_dma_buf_alloc(sc);
							}
						break;
					default:
						device_printf(sc->devicet, "unknown read mode (%08x)\n", (int)*mode);
						sc->error = TDS_ERROR_NOT_SUPPORTED;
						rv = ENOTTY;
						break;
					}
				}
			else {
				device_printf(sc->devicet, "can't set read mode while sampling in progress\n");
				sc->error = TDS_ERROR_READ_BUSY;
				rv = EBUSY;
				}
			break;
			}
		case TDSIO_GET_READ_MODE: {
			int *mode = (int *)data;
			*mode = sc->read_mode;
			break;
			}
		case TDSIO_GET_INPUT_STATUS: {
			u_int8_t input;
			int *status = (int *)data;
			*status = 0;
			input = tdsreg_get_input_status(sc);
			*status += input;
			break;
			}
		case TDSIO_GET_DMA_STATUS: {
			u_int8_t val;
			int *status = (int *)data;
			*status = 0;
			val = tdsreg_get_dma_status(sc);
			*status += val;
			break;
			}
		default:
			device_printf(sc->devicet, "unknown ioctl() command (%08x)\n", (int)cmd);
			sc->error = TDS_ERROR_NOT_SUPPORTED;
			rv = ENOTTY;
			break;
		}

	return rv;
}
